//############################################################################
// ~ International Laser Display Association ~
// ~     ILDA file format driver in C++      ~
// ~             built for EZFB              ~
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#include "ez_ilda.hpp"
#include "ez_wave.hpp"

//############################################################################
ez_ilda*     ez_ilda_pointer = NULL; // the one and only!
SDL_Surface* screen          = NULL; // an SDL graphics window
SDL_Event    event                 ;

//############################################################################
ez_ilda::ez_ilda(u_int xres, u_int yres)
          : ez_ilda_frame_set       (          ),
            ez_palette_set          (          ),
            bmp                     (          ),
            background              (          ),
            bg_file_name            ("."       ),
            bg_color                (0, 0, 0   ), // black
            wave_sample_rate        (48000     ), // Hz
            lit_dwell_overhang      (2         ),
            font_size_factor        (1         ),
            move_points_per_key     (100       ), // pxls
            output_bmp_size         (1000      ), // pxls square
            view_angle              (          ),
            view_scale              (1.0       ),
            view_offset             (          ),
            distance_delta_max      (600       ), // points
            longest_dwell_microsec  (333       ),
            insignificant_distance  (60        ), // points
            zero_angle_tolerance    (one_degree), // radians
            recolor_span_factor     (2.0       ),
            rendered_arc_angle      (6.0       ), // degrees
            frames_per_second       (16        ), // 0 == every frame is shown exactly once
            still_frame_duration    (10.0      ), // seconds
            show_points             (false     ),
            show_blanking           (false     ),
            show_cursors            (false     ),
            show_palettes           (true      ),
            show_stats              (true      ),
            show_fixed_bounds       (true      ),
            show_fixed_origin       (false     ),
            show_floating_axis      (false     ),
            show_floating_bounds    (false     ),
            show_ui_message         (true      ),
            show_arbitrary_point    (false     ),
            show_effects_generation (true      ),
            auto_scale_dxf          (true      ),
            maintain_dxf_origin     (false     ),
            loop_animation          (true      ),
            approximate_frame_rate  (true      ),
            black_dwell_points      (true      ),
            auto_flatten_z          (true      ),
            auto_optimize           (true      ),
            wave_show_points        (false     ),
            wave_show_blanks        (false     ),
            wave_real_time          (true      ),
            wave_show_inverted      (false     ),
            wave_clean_screen       (true      ),
            process_exception       (false     )
{
    if(ez_ilda_pointer == NULL)
    {
        ez_ilda_pointer = this;        
        wave_offsets = new int[6];        
        wave_offsets[0] = 0;
        wave_offsets[1] = 0;
        wave_offsets[2] = 0;
        wave_offsets[3] = 0;
        wave_offsets[4] = 0;
        wave_offsets[5] = 0;        
        load_cfg_file();
        if(SDL_Init(SDL_INIT_VIDEO) < 0)
            SDL_bailout(2);
        if((screen = SDL_SetVideoMode( xres,
                                       yres,
                                       0,
                                         SDL_HWSURFACE
                                       | SDL_ANYFORMAT
                                     )
           ) == NULL
          )
            SDL_bailout(2);
        SDL_EnableUNICODE(1);
        SDL_EnableKeyRepeat(250, 1);        
        font_size_factor = screen->h / 480;
        if(font_size_factor == 0)
            font_size_factor = 1;
        SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);
        SDL_EventState(SDL_PRESSED    , SDL_IGNORE);
        SDL_EventState(SDL_RELEASED   , SDL_IGNORE);
        SDL_EventState(SDL_SYSWMEVENT , SDL_IGNORE);
        SDL_WM_SetCaption("LaserBoy Multiplatform by James Lehman : Extra Stimulus Inc.", NULL);
        if(SDL_LockSurface(screen) < 0)
            SDL_bailout(2);        
        bmp_clone_SDL_Surface(&bmp, screen, 0);
        bmp_copy(&background, &bmp);
        recolor_background();
        bmp_to_SDL_Surface(&bmp, screen);
        load_effects();
    }
}

//############################################################################
ez_ilda::~ez_ilda()
{
    delete          wave_offsets;
    bmp_free        (&bmp);
    bmp_free        (&background);
    ez_ilda_pointer = NULL;
    SDL_FreeSurface (screen);
    SDL_Quit        ();
}

//############################################################################
void ez_ilda::sync_rgb_and_palette()
{
    int             i                         = 0          ,
                    j                                      ,
                    super_palette_first_frame = 0          ,
                    super_palette_last_frame  = 0          ,
                    current_frame_index       = frame_index;
    ez_palette_set  reduced_palette_set                    ;
    ez_palette      super_palette                          ;
    char            super_palette_name[7]                  ;
    static int      super_palette_number      = 0          ;
    //------------------------------------------------------------------------
    for(j = 0; j < number_of_frames(); j++)
        frame(j).sync_rgb_and_palette();
    //------------------------------------------------------------------------
    while(i < number_of_frames())
    {
        if(frame_picker(i).palette_index != EZ_TRUE_COLOR)
        {
            //--------------------------------------------------------------------
            super_palette_first_frame = i;
            super_palette = palette_picker(frame_picker(i).palette_index);
            sprintf(super_palette_name, "%06d", super_palette_number);
            super_palette.name = string("SP") + super_palette_name;
            super_palette_number++;
            //--------------------------------------------------------------------
            while(    i < number_of_frames()
                   && super_palette.unite(palette_picker(frame_picker(i).palette_index))
                 )
            i++;
            //--------------------------------------------------------------------
            reduced_palette_set.push_back(super_palette);
            super_palette_last_frame  = i - 1;
            //--------------------------------------------------------------------
            for(j = super_palette_first_frame; j <= super_palette_last_frame; j++)
            {
                frame(j).find_rgb_in_palette(super_palette);
                frame(j).palette_index = reduced_palette_set.number_of_palettes() - 1;
            }
            //--------------------------------------------------------------------
        }
        else
            i++;
    }
    //------------------------------------------------------------------------
    frame(current_frame_index);
    replace_palette_set(reduced_palette_set);
    palette_index = current_frame().palette_index;
    return;
}

//############################################################################
void ez_ilda::load_cfg_file()
{
    ifstream in(LASERBOY_CFG_FILE, ios::in);
    string   version;
    if(in.is_open())
    {
        int c;
        in >> version;
        if(version == LASERBOY_CFG_VERSION)
        {
            in >> bg_file_name           ;
            in >> c                      ;
            bg_color.r = c               ;
            in >> c                      ;
            bg_color.g = c               ;
            in >> c                      ;
            bg_color.b = c               ;
            in >> wave_sample_rate       ;
            in >> lit_dwell_overhang     ;
            in >> font_size_factor       ;
            in >> move_points_per_key    ;
            in >> output_bmp_size        ;
            in >> arbitrary_point.x      ;
            in >> arbitrary_point.y      ;
            in >> arbitrary_point.z      ;
            in >> view_angle.x           ;
            in >> view_angle.y           ;
            in >> view_angle.z           ;
            in >> view_scale.x           ;
            in >> view_scale.y           ;
            in >> view_scale.z           ;
            in >> view_offset.x          ;
            in >> view_offset.y          ;
            in >> view_offset.z          ;
            in >> distance_delta_max     ;
            in >> longest_dwell_microsec ;
            in >> insignificant_distance ;
            in >> zero_angle_tolerance   ;
            in >> recolor_span_factor    ;
            in >> rendered_arc_angle     ;
            in >> frames_per_second      ;
            in >> still_frame_duration   ;
            in >> show_points            ;
            in >> show_blanking          ;
            in >> show_cursors           ;
            in >> show_palettes          ;
            in >> show_stats             ;
            in >> show_fixed_bounds      ;
            in >> show_fixed_origin      ;
            in >> show_floating_axis     ;
            in >> show_floating_bounds   ;
            in >> show_effects_generation;
            in >> auto_scale_dxf         ;
            in >> maintain_dxf_origin    ;
            in >> loop_animation         ;
            in >> approximate_frame_rate ;
            in >> black_dwell_points     ;
            in >> auto_flatten_z         ;
            in >> auto_optimize          ;
            in >> wave_show_points       ;
            in >> wave_show_blanks       ;
            in >> wave_real_time         ;
            in >> wave_show_inverted     ;
            in >> wave_clean_screen      ;
            in >> wave_offsets[0]        ;
            in >> wave_offsets[1]        ;
            in >> wave_offsets[2]        ;
            in >> wave_offsets[3]        ;
            in >> wave_offsets[4]        ;
            in >> wave_offsets[5]        ;
        }
        in.close();
    }
    return;
}

//############################################################################
void ez_ilda::save_cfg_file()
{
    ofstream out(LASERBOY_CFG_FILE, ios::out);
    if(out.is_open())
    {
        out << LASERBOY_CFG_VERSION    << endl;
        out << bg_file_name.c_str()    << endl;
        out << (int)(bg_color.r)       << endl;
        out << (int)(bg_color.g)       << endl;
        out << (int)(bg_color.b)       << endl;
        out << wave_sample_rate        << endl;
        out << lit_dwell_overhang      << endl;
        out << font_size_factor        << endl;
        out << move_points_per_key     << endl;
        out << output_bmp_size         << endl;
        out << arbitrary_point.x       << endl;
        out << arbitrary_point.y       << endl;
        out << arbitrary_point.z       << endl;
        out << view_angle.x            << endl;
        out << view_angle.y            << endl;
        out << view_angle.z            << endl;
        out << view_scale.x            << endl;
        out << view_scale.y            << endl;
        out << view_scale.z            << endl;
        out << view_offset.x           << endl;
        out << view_offset.y           << endl;
        out << view_offset.z           << endl;
        out << distance_delta_max      << endl;
        out << longest_dwell_microsec  << endl;
        out << insignificant_distance  << endl;
        out << zero_angle_tolerance    << endl;
        out << recolor_span_factor     << endl;
        out << rendered_arc_angle      << endl;
        out << frames_per_second       << endl;
        out << still_frame_duration    << endl;
        out << show_points             << endl;
        out << show_blanking           << endl;
        out << show_cursors            << endl;
        out << show_palettes           << endl;
        out << show_stats              << endl;
        out << show_fixed_bounds       << endl;
        out << show_fixed_origin       << endl;
        out << show_floating_axis      << endl;
        out << show_floating_bounds    << endl;
        out << show_effects_generation << endl;
        out << auto_scale_dxf          << endl;
        out << maintain_dxf_origin     << endl;
        out << loop_animation          << endl;
        out << approximate_frame_rate  << endl;
        out << black_dwell_points      << endl;
        out << auto_flatten_z          << endl;
        out << auto_optimize           << endl;
        out << wave_show_points        << endl;
        out << wave_show_blanks        << endl;
        out << wave_real_time          << endl;
        out << wave_show_inverted      << endl;
        out << wave_clean_screen       << endl;    
        out << wave_offsets[0]         << endl;
        out << wave_offsets[1]         << endl;
        out << wave_offsets[2]         << endl;
        out << wave_offsets[3]         << endl;
        out << wave_offsets[4]         << endl;
        out << wave_offsets[5]         << endl << endl;
        out << "If this file is currupted, delete it!" << endl;
        out.close();
    }
    return;
}

//############################################################################
void ez_ilda::display(string ui_message)
{
    if(palette_index != current_frame().palette_index)
        palette_index = current_frame().palette_index;
    current_frame().render_in_ezfb_bitmap(ui_message);
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
string ez_ilda::display_prompt_and_echo_name(string prompt)
{
    bool   done = false;

    int    x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off  = BMP.yres - 24 * font_size_factor,
           x_gap  = 8 * font_size_factor,
           y_line = x_gap;

    string input;

    process_exception = false;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(0,255,255)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "_"
              );
    bmp_to_SDL_Surface(&bmp, screen);

    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);
        if(event.key.keysym.sym == SDLK_ESCAPE)
        {
            process_exception = true;
            return "";
        }
        if(event.key.keysym.sym == SDLK_RETURN)
            done = true;
        else
        {
            switch(event.key.keysym.sym)
            {
                //-------------------------------------------------    
                case SDLK_LEFT:
                case SDLK_DELETE:
                case SDLK_BACKSPACE:
                    if(input.size())
                    {
                        input = input.substr(0, input.size() - 1);
                        x_off -= x_gap;
                        bmp_printf(   &BMP
                                    , x_off
                                    , y_off
                                    , BMP.make_rgb(0,255,255)
                                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                                    , 0
                                    , font_size_factor
                                    , " "
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                    }                         
                    break;
                //-------------------------------------------------    
                default:
                    if(    isalnum(event.key.keysym.sym)
                        || event.key.keysym.unicode == '_'
                      )
                    {
                        bmp_printf(   &BMP
                                    , x_off
                                    , y_off
                                    , BMP.make_rgb(0,255,255)
                                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                                    , 0
                                    , font_size_factor
                                    , "%c"
                                    , event.key.keysym.unicode
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                        input += event.key.keysym.unicode;
                        x_off += x_gap;
                    }                            
                    break;                         
                //-------------------------------------------------    
            } // end switch(event.key.keysym.sym)
        } // end else of if(event.key.keysym.sym == SDLK_RETURN)
    } // end while(!done)    
    bmp_printf(   &BMP
                , x_off - x_gap * input.size()
                , y_off
                , BMP.make_rgb(255,0,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , input.c_str()
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return input;
}

//############################################################################
string ez_ilda::display_prompt_and_echo_string(string prompt)
{
    bool   done = false;

    int    x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off  = BMP.yres - 24 * font_size_factor,
           x_gap  = 8 * font_size_factor,
           y_line = 8 * font_size_factor;

    string input;

    process_exception = false;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(0,255,255)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "_"
              );
    bmp_to_SDL_Surface(&bmp, screen);
              
    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(    event.type != SDL_KEYDOWN
//                 && event.type != SDL_QUIT
               );               
//        if(event.type == SDL_QUIT)
//        {
//            process_exception = true;
//            return "";
//        }
        if(event.key.keysym.sym == SDLK_ESCAPE)
        {
            process_exception = true;
            return "";
        }
        if(event.key.keysym.sym == SDLK_RETURN)
            done = true;
        else
        {
            switch(event.key.keysym.sym)
            {
                //-------------------------------------------------    
                case SDLK_LEFT:
                case SDLK_DELETE:
                case SDLK_BACKSPACE:
                    if(input.size())
                    {
                        input = input.substr(0, input.size() - 1);
                        x_off -= x_gap;
                        bmp_printf(   &BMP
                                    , x_off
                                    , y_off
                                    , BMP.make_rgb(0,255,255)
                                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                                    , 0
                                    , font_size_factor
                                    , " "
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                    }                         
                    break;
                //-------------------------------------------------    
                default:
                    if(    isgraph(event.key.keysym.sym)
                        || event.key.keysym.sym == ' '
                      )
                    {
                        bmp_printf(   &BMP
                                    , x_off
                                    , y_off
                                    , BMP.make_rgb(0,255,255)
                                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                                    , 0
                                    , font_size_factor
                                    , "%c"
                                    , event.key.keysym.unicode
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                        input += event.key.keysym.unicode;
                        x_off += x_gap;
                    }                            
                    break;                         
                //-------------------------------------------------    
            } // end switch(event.key.keysym.sym)
        } // end else of if(event.key.keysym.sym == SDLK_RETURN)
    } // end while(!done)
    bmp_printf(   &BMP
                , x_off - x_gap * input.size()
                , y_off
                , BMP.make_rgb(255,0,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , input.c_str()
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return input;
}

//############################################################################
float ez_ilda::display_prompt_and_echo_float(string prompt)
{
    bool   done = false;

    int    x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off  = BMP.yres - 24 * font_size_factor,
           x_gap  = 8 * font_size_factor,
           y_line = 8 * font_size_factor;

    float  result;

    string input;

    process_exception = false;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(0,255,255)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "_"
              );
    bmp_to_SDL_Surface(&bmp, screen);

    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(    event.type != SDL_KEYDOWN
//                 && event.type != SDL_QUIT
               );               
//        if(event.type == SDL_QUIT)
//        {
//            process_exception = true;
//            return 0.0;
//        }
        if(event.key.keysym.sym == SDLK_ESCAPE)
        {
            process_exception = true;
            return 0.0;
        }
        if(event.key.keysym.sym == SDLK_RETURN)
            done = true;
        else
        {
            switch(event.key.keysym.sym)
            {
                //-------------------------------------------------    
                case SDLK_LEFT:
                case SDLK_DELETE:
                case SDLK_BACKSPACE:
                    if(input.size())
                    {
                        input = input.substr(0, input.size() - 1);
                        x_off -= x_gap;
                        bmp_printf(   &BMP
                                    , x_off
                                    , y_off
                                    , BMP.make_rgb(0,255,255)
                                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                                    , 0
                                    , font_size_factor
                                    , " "
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                    }                         
                    break;
                //-------------------------------------------------    
                default:
                    if(    isdigit(event.key.keysym.sym)
                        || (    event.key.keysym.sym == '.'
                             && input.find('.') == string::npos
                           )
                        || (    event.key.keysym.sym == '+'
                             && !input.size()
                           )
                        || (    event.key.keysym.sym == '-'
                             && !input.size()
                           )
                      )                            
                    {
                        bmp_printf(   &BMP
                                    , x_off
                                    , y_off
                                    , BMP.make_rgb(0,255,255)
                                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                                    , 0
                                    , font_size_factor
                                    , "%c"
                                    , event.key.keysym.sym
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                        input += event.key.keysym.sym;
                        x_off += x_gap;
                    }                            
                    break;                         
                //-------------------------------------------------    
            } // end switch(event.key.keysym.sym)
        } // end else of if(event.key.keysym.sym == SDLK_RETURN)
    } // end while(!done)    
    bmp_printf(   &BMP
                , x_off - x_gap * input.size()
                , y_off
                , BMP.make_rgb(255,0,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , input.c_str()
              );
    bmp_to_SDL_Surface(&bmp, screen);
    sscanf(input.c_str(), "%f", &result);
    return result;
}

//############################################################################
int ez_ilda::display_prompt_and_echo_int(string prompt)
{
    bool   done = false;
    
    int    x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off  = BMP.yres - 24 * font_size_factor,
           x_gap  = 8 * font_size_factor,
           y_line = 8 * font_size_factor,
           result;

    string input;
    
    process_exception = false;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(0,255,255)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "_"
              );
    bmp_to_SDL_Surface(&bmp, screen);

    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(    event.type != SDL_KEYDOWN
//                 && event.type != SDL_QUIT
               );
//        if(event.type == SDL_QUIT) // clicked window X
//        {
//            process_exception = true;
//            return 0;
//        }
        if(event.key.keysym.sym == SDLK_ESCAPE)
        {
            process_exception = true;
            return 0;
        }
        if(event.key.keysym.sym == SDLK_RETURN)
            done = true;
        else
        {
            switch(event.key.keysym.sym)
            {
                //-------------------------------------------------    
                case SDLK_LEFT:
                case SDLK_DELETE:
                case SDLK_BACKSPACE:
                    if(input.size())
                    {
                        input = input.substr(0, input.size() - 1);
                        x_off -= x_gap;
                        bmp_printf(   &BMP
                                    , x_off
                                    , y_off
                                    , BMP.make_rgb(0,255,255)
                                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                                    , 0
                                    , font_size_factor
                                    , " "
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                    }                         
                    break;
                //-------------------------------------------------    
                default:
                    if(    isdigit(event.key.keysym.sym)
                        || (    event.key.keysym.sym == '+'
                             && !input.size()
                           )
                        || (    event.key.keysym.sym == '-'
                             && !input.size()
                           )
                      )                    
                    {
                        bmp_printf(   &BMP
                                    , x_off
                                    , y_off
                                    , BMP.make_rgb(0,255,255)
                                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                                    , 0
                                    , font_size_factor
                                    , "%c"
                                    , event.key.keysym.sym
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                        input += event.key.keysym.sym;
                        x_off += x_gap;
                    }                            
                    break;                         
                //-------------------------------------------------    
            } // end switch(event.key.keysym.sym)
        } // end else of if(event.key.keysym.sym == SDLK_RETURN)
    } // end while(!done)
    bmp_printf(   &BMP
                , x_off - x_gap * input.size()
                , y_off
                , BMP.make_rgb(255,0,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , input.c_str()
              );
    bmp_to_SDL_Surface(&bmp, screen);
    if(input.size())
        sscanf(input.c_str(), "%d", &result);
    else
        result = 0;
    return result;
}

//############################################################################
bool ez_ilda::display_prompt_and_echo_bool(string prompt)
{
    int    x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off  = BMP.yres - 24 * font_size_factor;

    process_exception = false;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , prompt.c_str()
              );
    bmp_to_SDL_Surface(&bmp, screen);    
    do
    {
        SDL_WaitEvent(&event);
    } while(    event.type != SDL_KEYDOWN
//             && event.type != SDL_QUIT
           );
//    if(event.type == SDL_QUIT) // clicked window X
//    {
//        process_exception = true;
//        return false;
//    }
    if(event.key.keysym.sym == SDLK_ESCAPE)
    {
        process_exception = true;
        return false;
    }
    if(event.key.keysym.sym == 'y')
        return true;
    return false;
}

//############################################################################
void ez_ilda::display_program_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "Welcome to " LASERBOY_VERSION " !!!"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      i open file in"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      o save file out"       
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      u user interface visuals"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      v view keys & apply"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    ` ~ auto show"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    - _ auto orbit"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    = + auto show & orbit"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    %c %c frame"
                , 0x1b
                , 0x1a
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    %c %c target palette"
                , 0x18
                , 0x19
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      9 first frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      0 last frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      ( set | unset frames-( select"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      ) set | unset )-frames select"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Space] select | unselect frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Bkspc] delete current frame"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      p palette   transforms"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      h frame     transforms"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      j frame set transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      k draw, color, blank  (vertices)"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      l move, scale, rotate (segments)"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      b LaserBoy wave utilities"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    / ? frame set stats | frame stats"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "  [Tab] system settings"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "  [Esc] quit saving OUT.ild"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Enter] toggle this help"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_please_wait()
{
    int x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off,
        y_line = 8 * font_size_factor;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "please wait..."
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_in_file_type_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "file input type"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    1 ILDA"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    2 DXF"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    3 LaserBoy wave"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    4 ADAT wave set"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    5 bitmap"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    6 ascii palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Esc] cancel action"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_in_ild_action_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "ILDA file input action"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    1 replace frame set"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    2 prepend to frame set"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    3 insert after current frame"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    4 append to frame set"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    5 superimpose frame(s) into frame set"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Esc] cancel action"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_in_dxf_action_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "DXF file input action"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    1 replace frame set with DXF file"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    2 append  frame set with DXF file"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    3 replace frame set with DXF directory"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    4 append  frame set with DXF directory"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Esc] cancel action"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_in_wave_action_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "wave file input action"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    1 replace current frame set"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    2 append to current frame set"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Esc] cancel action"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_in_bmp_action_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "bitmap file input action"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    1 open as frame set color map"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    2 open into frame background"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    3 clear frame background"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Esc] cancel action"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
//void ez_ilda::display_in_pal_action_menu()
//{
//     return;
//}

//############################################################################
void ez_ilda::display_out_file_type_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "Output file type"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    1 ILDA save current frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    2 ILDA save selected frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    3 ILDA save all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    4 ILDA generate & save frame effect"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    5 ILDA generate & save frame set effect"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    6 DXF save current frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    7 DXF save directory selected frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    8 DXF save directory all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    9 wave animation %3.3f fps of all frames"
                , frames_per_second
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    0 wave %6.2f second still of current frame"
                , still_frame_duration
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    a C++ save current frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    b C++ save selected frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    c C++ save all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    d bmp save %d%c pxl of current frame"
                , output_bmp_size
                , 0xfd
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    e bmp save directory %d%c pxl selected frames"
                , output_bmp_size
                , 0xfd
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    f bmp save directory %d%c pxl all frames"
                , output_bmp_size
                , 0xfd
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    g ASCII palette save current frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    h ASCII palette save target palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    i save system settings [LaserBoy.wtf]"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[ESC] cancel action"
              );
    bmp_to_SDL_Surface(&bmp, screen);    
    return;
}

//############################################################################
void ez_ilda::display_settings_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "LaserBoy settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    1 insignificant angle in degrees ..... %6.2f%c"
                , zero_angle_tolerance * one_radian
                , 0xf8                
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    2 insignificant distance in points ... %6.2f"
                , insignificant_distance
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    3 longest optimized vector in points . %6.2f"
                , distance_delta_max
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    4 maximum dwell in microseconds ...... %6.2f"
                , longest_dwell_microsec
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    5 lit dwell overhang in samples ...... %d"
                , lit_dwell_overhang
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    6 DXF arc step in degrees ............ %6.2f%c"
                , rendered_arc_angle
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    7 move points per key hit ............ %d"
                , move_points_per_key
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    8 recolor palette span factor ........ %6.2f"
                , recolor_span_factor
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    9 animation frames per second ........ %6.2f"
                , frames_per_second
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    0 still frame durration in seconds ... %6.2f"
                , still_frame_duration
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    a output wave sample rate ............ %d"
                , wave_sample_rate
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    b output bitmap size in square pixels  %d"
                , output_bmp_size
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    c display font size factor ........... %d"
                , font_size_factor
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    d display background color ....... red %d"
                , bg_color.r
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    e display background color ..... green %d"
                , bg_color.g
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    f display background color ...... blue %d"
                , bg_color.b
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[ESC] exit settings menu"
              );
    bmp_to_SDL_Surface(&bmp, screen);    
    return;
}

//############################################################################
void ez_ilda::display_veiw_menu()
{
    int x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off,
        y_line =  8 * font_size_factor,
        x_tab  = 17 * y_line,
        x_tab2 =  9 * y_line;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "press & hold view keys, [Shift] reverse"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "1 ! orbit X"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,255)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "5 %% front | back"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "2 @ orbit Y"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,255)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "6 ^  left | right"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "3 # orbit Z"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,255)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "7 &   top | bottom"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "4 $ orbit XYZ"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,255)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "8 * orbit & frame"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "q Q pan X"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "a A zoom X"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "w W pan Y"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "s S zoom Y"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "e E pan Z"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "d D zoom Z"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "r R pan XYZ"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "f F zoom XYZ"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "t T center"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "g G 1:1:1"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "current view"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_tab2
                , y_off
                , BMP.make_rgb(255,0,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "X"
              );
    bmp_printf(   &BMP
                , x_tab2 * 2
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "Y"
              );
    bmp_printf(   &BMP
                , x_tab2 * 3
                , y_off
                , BMP.make_rgb(0,0,255)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "Z"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , " angle                               degrees"
              );
    bmp_printf(   &BMP
                , x_tab2
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.make_rgb(80,0,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_angle.x * one_radian
              );
    bmp_printf(   &BMP
                , x_tab2 * 2
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.make_rgb(0,80,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_angle.y * one_radian
              );
    bmp_printf(   &BMP
                , x_tab2 * 3
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.make_rgb(0,0,80)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_angle.z * one_radian
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "offset                               points"
              );
    bmp_printf(   &BMP
                , x_tab2
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.make_rgb(80,0,0)
                , 0
                , font_size_factor
                , "%-8.0f"
                , view_offset.x
              );
    bmp_printf(   &BMP
                , x_tab2 * 2
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.make_rgb(0,80,0)
                , 0
                , font_size_factor
                , "%-8.0f"
                , view_offset.y
              );
    bmp_printf(   &BMP
                , x_tab2 * 3
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.make_rgb(0,0,80)
                , 0
                , font_size_factor
                , "%-8.0f"
                , view_offset.z
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , " scale"
              );
    bmp_printf(   &BMP
                , x_tab2
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.make_rgb(80,0,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_scale.x
              );
    bmp_printf(   &BMP
                , x_tab2 * 2
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.make_rgb(0,80,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_scale.y
              );
    bmp_printf(   &BMP
                , x_tab2 * 3
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.make_rgb(0,0,80)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_scale.z
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      1 apply angle  to frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      2 apply offset to frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      3 apply scale  to frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Shift] 1, 2, 3 apply to all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "  [Esc] exit view settings"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_visuals_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "LaserBoy toggle UI visuals settings"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((show_points)?("--> 1 vertices"):("    1 vertices"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((show_blanking)?("--> 2 blanking"):("    2 blanking"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((show_palettes)?("--> 3 palettes"):("    3 palettes"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((show_stats)?("--> 4 stats"):("    4 stats"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((show_fixed_bounds)?("--> 5 fixed bounds"):("    5 fixed bounds"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((show_fixed_origin)?("--> 6 fixed origin"):("    6 fixed origin"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((show_floating_axis)?("--> 7 floating axis"):("    7 floating axis"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((show_floating_bounds)?("--> 8 floating bounds"):("    8 floating bounds"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((show_ui_message)?("--> 9 UI message"):("    9 UI message"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    0 show | hide all visuals"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "toggle LaserBoy system settings"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((show_effects_generation)?("--> a show effect rendering"):("    a show effect rendering"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((auto_flatten_z)?("--> b flatten Z for wave output"):("    b flatten Z for wave output"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((auto_optimize)?("--> c optimize before save"):("    c optimize before save"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((loop_animation)?("--> d loop animation"):("    d loop animation"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((approximate_frame_rate)?("--> e animate at approximated frame rate"):("    e animate at approximated frame rate"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((auto_scale_dxf)?("--> f auto scale DXF import"):("    f auto scale DXF import"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((maintain_dxf_origin)?("--> g maintain DXF origin"):("    g maintain DXF origin"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((black_dwell_points)?("--> h black dwell"):("    h black dwell"))
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Esc] exit LaserBoy UI visuals settings"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_palette_tranz_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "palette transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      1 span X with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      2 span Y with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      3 span Z with palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      4 span  vectors with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      5 index vectors with palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      6 span  lit segments with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      7 index lit segments with palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      8 random color vectors with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      9 random color lit segments with palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      0 set to target palette by index"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      a best match to target palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      b convert true color to palette (if possible)"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      c bit reduce true color to palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      d promote to true color"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Shift] select to apply to all frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "  [Esc] exit palette transforms"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_frame_set_tranz_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "frame set transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      1 delete selected"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      2 trim to selected"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      3 delete repeats from wave import"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      4 copy selected to begining"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      5 copy selected to end"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      6 move selected to begining"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      7 move selected to end"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      8 add empty frame to begining"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      9 add empty frame after current"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      0 add empty frame to end"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      a reverse frame set order"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      > explode frame into segment set"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      < collapse selected frames into one"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "  [Esc] exit frame set transforms"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_frame_tranz_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "frame transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      1 flatten Z (force to 2D)"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      2 Z order vertices"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      3 reverse vector order"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      4 mirror"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      5 flip"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      6 quarter turn CW"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      7 quarter turn CCW"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      8 accentuate the positive"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      9 eliminate the negative"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      0 remove blank span"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      a remove angle dwell"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      b reduce vectors"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      c bond segments (reorder)"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      d minimize vertices (0,a,b,c)"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      e fracture segments"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      f reorder segments"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      g add angle dwell"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      h add distance span"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      i optimize"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Shift] select applies to all frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "  [Esc] exit frame transforms"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_draw_color_blank_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "draw, color, blank mode entered"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    \\ | %c to 0 | %c to end"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    [ { +- %c   one vertex"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    ] } +-   %c one vertex"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    i I +- %c   lit segment"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    o O +-   %c lit segment"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    u U +- %c-%c lit segment"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    a A unblank all vectors | frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    b B blank | unblank %c-%c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    p P +- palette color index"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      c color %c-%c with palette index"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      C get palette index from %c color"
                , 0x01
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    x X move   %c X +- 100"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    y Y move   %c Y +- 100"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    z Z move   %c Z +- 100"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      q set    %c X"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      w set    %c Y"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      e set    %c Z"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    f F rotate %c +- 1%c on Z axis"
                , 0x01
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    g G rotate %c +- 1%c on Y axis"
                , 0x01
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    m M +- 1%% magnitude of %c"
                , 0x01
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      l draw line %c to %c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      t render text on line %c to %c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      d %3.2f%c circular  arc, %c to %c radius"
                , rendered_arc_angle
                , 0xf8
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      D %3.2f%c eliptical arc, %c to %c radius"
                , rendered_arc_angle
                , 0xf8
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      k draw polygon %c to %c radius"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      . split vector in line or add vector to end"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      s add blank vertex at %c (cut segment)"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Bkspc] delete vertex at %c"
                , 0x01
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    / ? show frame | set stats"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "%c %c %c %c and view orbit keys are active"
                , 0x1b
                , 0x1a
                , 0x18
                , 0x19
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Enter] +- this help"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "  [Esc] exit draw, color blank mode"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_move_scale_rotate_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "move, scale, rotate mode entered"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    \\ | %c to 0 | %c to end"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    [ { +- %c   one vertex"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    ] } +-   %c one vertex"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    i I +- %c   lit segment"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    o O +-   %c lit segment"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    u U +- %c-%c lit segment"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    x X move   %c-%c X +- 100"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    y Y move   %c-%c Y +- 100"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    z Z move   %c-%c Z +- 100"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      q set    %c-%c center X"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      w set    %c-%c center Y"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      e set    %c-%c center Z"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    f F rotate %c-%c +- 1%c Z axis on center"
                , 0x01
                , 0x0f
                , 0xf8                
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    g G rotate %c-%c +- 1%c Y axis on center"
                , 0x01
                , 0x0f
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    9 ( scale  %c-%c X      +- 1%% on center"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    0 ) scale  %c-%c  Y     +- 1%% on center"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    - _ scale  %c-%c   Z    +- 1%% on center"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    = + scale  %c-%c XYZ    +- 1%% on center"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    a A move   a point X +- 100"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    s S move   a point Y +- 100"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    d D move   a point Z +- 100"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      j set    a point X"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      k set    a point Y"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      l set    a point Z"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    r R rotate %c-%c +- 1%c Z axis on a point"
                , 0x01
                , 0x0f
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    t T rotate %c-%c +- 1%c Y axis on a point"
                , 0x01
                , 0x0f
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    v V scale  %c-%c X    +- 1%% from a point"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    b B scale  %c-%c  Y   +- 1%% from a point"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    n N scale  %c-%c   Z  +- 1%% from a point"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    m M scale  %c-%c XYZ  +- 1%% from a point"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "      c copy   %c-%c (in place) to end"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    / ? frame | set stats"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "%c %c %c %c and view orbit keys are active"
                , 0x1b
                , 0x1a
                , 0x18
                , 0x19
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Enter] toggle help"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "  [ESC] exit move, scale, rotate mode"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_wave_utilities_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "LaserBoy wave utilities"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    1 LaserBoy wave show settings"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    2 show LaserBoy wave stats"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    3 show LaserBoy wave on screen [buggy!]"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    4 set LaserBoy wave sample offsets"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    5 apply offsets to LaserBoy wave"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "    6 invert polarity of LaserBoy wave"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[ESC] exit LaserBoy wave utilities"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_wave_show_settings_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "LaserBoy wave show settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((wave_show_points)?("--> 1 vertices"):("    1 vertices"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((wave_show_blanks)?("--> 2 blanking"):("    2 blanking"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((wave_real_time)?("--> 3 real time"):("    3 real time"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((wave_show_inverted)?("--> 4 inverted"):("    4 inverted"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((wave_clean_screen)?("--> 5 clean screen"):("    5 clean screen"))
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Esc] exit LaserBoy wave show settings"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_wave_stats(string file_name)
{
    int            i,
                   y_line = 8 * font_size_factor,
                   x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
                   y_off  = x_off;        
    ifstream       in;
    ez_wave_header header;

    process_exception = false;
    
    if(    file_name.size() < 4
        || file_name.substr(file_name.size() - 4) != ".wav"
      )
        file_name += ".wav";
    in.open((EZ_ILDA_WAVE_SHARE + file_name).c_str(), ios::in | ios::binary);
    //------------------------------------------------------------------------
    if(!in.is_open())
    {
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , BMP.white
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "[any key] %s failed to open"
                    , file_name.c_str()
                  );
        error = EZ_ILDA_FILE_OPEN_FAILED;                  
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    error = header.from_ifstream(in);
    in.close();
    //------------------------------------------------------------------------
    if(error)
    {
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , BMP.white
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "[any key] %s wave header read failed"
                    , file_name.c_str()
                  );
        process_exception = true;
        bmp_to_SDL_Surface(&bmp, screen);
        return;
    }
    //------------------------------------------------------------------------
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "LaserBoy wave stats"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , (EZ_ILDA_WAVE_SHARE + file_name).c_str()
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "LaserBoy wave version : %s"
                , header.header_version.c_str()
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "num_samples ..........: %d"
                , header.num_samples
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "sample_rate ..........: %d"
                , header.sample_rate
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , ((header.ez_wave_mode & EZ_WAVE_POSITIVE) ? ("LaserBoy subcodes ....: positive signal") : ("LaserBoy subcodes ....: negative signal"))
              );              
    if(header.ez_wave_mode & EZ_WAVE_END_OF_FRAME)
    {
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "                      : end of frame marks"
                  );
    }
    if(header.ez_wave_mode & EZ_WAVE_UNIQUE_FRAME)
    {
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "                      : unique frame marks"
                  );
    }
    if(header.ez_wave_mode & EZ_WAVE_UNIQUE_POINT)
    {
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "                      : unique point marks"
                  );
    }
    if(header.ez_wave_mode & EZ_WAVE_OFFSETS)
    {
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "                      : with sample offsets"
                  );
    }
    else
    {
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "                      : no sample offsets"
                  );
    }
    if(header.ez_wave_mode & EZ_WAVE_OPTIMIZED)
    {
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "                      : optimized for galvos"
                  );
    }    
    else
    {
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "                      : not optimized"
                  );
    }    
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "number of channels ...: %d"
                , header.num_channels
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "bits per sample ......: %d"
                , header.bits_per_sample
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "playing time .........: %f seconds"
                , ((double)header.num_samples / header.sample_rate)
              );
    y_off += y_line;    
    for(i = 0; i < header.num_channels; i++)
    {
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "............ offset[%d]: %d"
                    , i
                    , header.offset[i]
                  );
    }
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "lit dwell overhang ...: %d"
                , header.parms.lit_dwell_overhang
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "distance delta max ...: %6.2f"
                , header.parms.distance_delta_max
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "longest dwell u_sec ..: %6.2f"
                , header.parms.longest_dwell_microsec
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "insignificant distance: %6.2f"
                , header.parms.insignificant_distance
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "insignificant angle ..: %6.2f"
                , header.parms.zero_angle_tolerance
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "frames per second ....: %6.2f"
                , header.parms.frames_per_second
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "press any key to continue"
                , ((double)header.num_samples / header.sample_rate)
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_wave(string file_name)
{
    bool                     playing      = true         ;
    int                      i                           ,
                             error                       ,
                             span                        ,
                             last                        ,
                             red                         ,
                             green                       ,
                             blue                        ,
                             sample_count = 0            ,
                             offset_max                  ,
                             offset_min                  ,
                             center       = screen->h / 2;
    u_int                    time_zero                   ,
                             color        = BMP.white    ;
    float                    scale  = screen->h / 65536.0;
    ifstream                 in                          ;
    ez_wave_header           header                      ;
    struct  ezfb_bitmap      frame_buffer = {0}          ;
    ez_wave_sample_pointer   roll_over                   ;
    ez_wave_sample_pointer*  sample_window               ;
    //------------------------------------------------------------------------
    process_exception = false;
    //------------------------------------------------------------------------
    do // clear key strokes from event queue!
    {
        SDL_PollEvent(&event);    
    } while(event.type == SDL_KEYDOWN);
    //------------------------------------------------------------------------
    if(    file_name.size() < 4
        || file_name.substr(file_name.size() - 4) != ".wav"
      )
        file_name += ".wav";
    in.open((EZ_ILDA_WAVE_SHARE + file_name).c_str(), ios::in | ios::binary);
    //------------------------------------------------------------------------
    if(!in.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , BMP.white
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "[any key] %s failed to open               "
                    , file_name.c_str()
                  );
        error = EZ_ILDA_FILE_OPEN_FAILED;
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    error = header.from_ifstream(in);
    if(error)
    {
        in.close();              
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , BMP.white
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "[any key] %s wave header read failed      "
                  );
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    if(header.num_channels < 6)
    {
        in.close();              
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , BMP.white
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "[any key] %s is not a 6 channel wave         "
                    , file_name.c_str()
                  );
        error = EZ_ILDA_INCOMPATIBLE_FORMATS;
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        return;
    }
    //------------------------------------------------------------------------    
    offset_min = header.offset[0];
    offset_max = header.offset[0];    
    for(i = 1; i < 6; i++)
    {
        if(header.offset[i] > offset_max)
            offset_max = header.offset[i];
        if(header.offset[i] < offset_min)
            offset_min = header.offset[i];
    }
    //--------------------------------------------------------------------
    span  = (offset_max - offset_min) + 2; // at least 2 samples!
    last  = span - 1;
    red   = -header.offset[2] + 1;
    green = -header.offset[3] + 1;
    blue  = -header.offset[4] + 1;
    //------------------------------------------------------------------------
    sample_window = new ez_wave_sample_pointer[span]; // at least 2
    for(i = 0; i < span; i++)
        sample_window[i] = new ez_wave_sample(header.num_channels);
    //------------------------------------------------------------------------    
    bmp_clone_SDL_Surface(&frame_buffer, screen, 0);
    bmp_uint(&frame_buffer, bg_color.as_uint());
    //------------------------------------------------------------------------    
    while(playing)
    {
        time_zero = SDL_GetTicks();
        //----------------------------------------------------------------
        roll_over = sample_window[last - 1];
        for(i = 0; i < last; i++)
            sample_window[i] = sample_window[i + 1];
        sample_window[last] = roll_over;
        //----------------------------------------------------------------
        if(wave_show_inverted)
        {
            if(sample_window[last]->from_ifstream_inverted(in)) // if EOF
                playing = false;
        }
        else
        {
            if(sample_window[last]->from_ifstream(in)) // if EOF
                playing = false;
        }
        //----------------------------------------------------------------
        sample_count++;        
        //----------------------------------------------------------------
        color = frame_buffer.make_rgb(   sample_window[red  ]->channel[2] >> 7
                                       , sample_window[green]->channel[3] >> 7
                                       , sample_window[blue ]->channel[4] >> 7
                                     );
        //----------------------------------------------------------------
        if(color)
        {
            bmp_put_line_fast(   &frame_buffer
                               , (u_int)((sample_window[0]->channel[0] *  scale) + center)
                               , (u_int)((sample_window[0]->channel[1] * -scale) + center)
                               , (u_int)((sample_window[1]->channel[0] *  scale) + center)
                               , (u_int)((sample_window[1]->channel[1] * -scale) + center)
                               , color
                             );

            if(wave_show_points)
                bmp_put_rectangle(   &frame_buffer
                                   , (u_int)((sample_window[1]->channel[0] *  scale) + center - 1)
                                   , (u_int)((sample_window[1]->channel[1] * -scale) + center - 1)
                                   , (u_int)((sample_window[1]->channel[0] *  scale) + center + 1)
                                   , (u_int)((sample_window[1]->channel[1] * -scale) + center + 1)
                                   , color
                                   , false
                                 );
        }
        //----------------------------------------------------------------
        else if(wave_show_blanks)
        {
            bmp_put_line_fast(   &frame_buffer
                               , (u_int)((sample_window[0]->channel[0] *  scale) + center)
                               , (u_int)((sample_window[0]->channel[1] * -scale) + center)
                               , (u_int)((sample_window[1]->channel[0] *  scale) + center)
                               , (u_int)((sample_window[1]->channel[1] * -scale) + center)
                               , frame_buffer.gray
                             );

            if(wave_show_points)
                bmp_put_rectangle(   &frame_buffer
                                   , (u_int)((sample_window[1]->channel[0] *  scale) + center - 1)
                                   , (u_int)((sample_window[1]->channel[1] * -scale) + center - 1)
                                   , (u_int)((sample_window[1]->channel[0] *  scale) + center + 1)
                                   , (u_int)((sample_window[1]->channel[1] * -scale) + center + 1)
                                   , frame_buffer.gray
                                   , false
                                  );
        }
        //----------------------------------------------------------------
        if(sample_window[red]->channel[2] & 0x0001)
        {
            if(wave_real_time)
            {
                do
                {    
                     SDL_Delay(1);
                }
                while(   SDL_GetTicks() - time_zero
                       < (u_int)(sample_count / (header.sample_rate / 1000.0))
                     );                             
            }
            bmp_to_SDL_Surface(&frame_buffer, screen);
            if(wave_clean_screen)
                bmp_uint(&frame_buffer, bg_color.as_uint());
            SDL_PollEvent(&event);
            if(event.type == SDL_KEYDOWN)
                playing = false;
            sample_count = 0;
        }
        //----------------------------------------------------------------
    } // end while(playing)
    in.close();
    bmp_free(&frame_buffer);
//    for(i = 0; i < span; i++)
//        delete sample_window[i];
    delete sample_window;
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_ilda::display_wave_offsets()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "LaserBoy wave sample offset settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "  [0] X          %d"
                , wave_offsets[0]
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "  [1] Y          %d"
                , wave_offsets[1]
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "  [2] red        %d"
                , wave_offsets[2]
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "  [3] green      %d"
                , wave_offsets[3]
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "  [4] blue       %d"
                , wave_offsets[4]
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "  [5] monochrome %d"
                , wave_offsets[5]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "[Esc] cancel"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_file_list(string directory)
{
    bool            found_one = false;
    int             i,
                    y_line = 8 * font_size_factor,
                    x_off  = BMP.yres + EZ_STATS_DISPLAY_OFFSET * font_size_factor,
                    y_off  = y_line;
    DIR            *pDIR;
    struct dirent  *pDirEnt;

    pDIR = opendir(directory.c_str());
    if(pDIR == NULL)
        return;
    bmp_printf(   &BMP
                , x_off
                , 0
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "Palette set stats "
              );
    pDirEnt = readdir(pDIR);  
    while(pDirEnt != NULL)
    {
        if(    strcmp(pDirEnt->d_name, ".")  // not the dot
            && strcmp(pDirEnt->d_name, "..") // not the dot dot
          )
        {
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , next_hue()
                        , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                        , 0
                        , font_size_factor
                        , "%s                  "
                        , pDirEnt->d_name
                      );
            y_off += y_line;
            found_one = true;
        }
        pDirEnt = readdir(pDIR);
    }
    closedir(pDIR);    
    if(!found_one)
    {
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "none found        "
                  );
        y_off += y_line;
    }    
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_directory_list(string directory)
{
    bool            found_one = false;
    int             i,
                    y_line = 8 * font_size_factor,
                    x_off  = BMP.yres + EZ_STATS_DISPLAY_OFFSET * font_size_factor,
                    y_off  = y_line;
    DIR            *pDIR;
    struct dirent  *pDirEnt;

    pDIR = opendir(directory.c_str());
    if(pDIR == NULL)
        return;
    bmp_printf(   &BMP
                , x_off
                , 0
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "Palette set stats "
              );
    pDirEnt = readdir(pDIR);  
    while(pDirEnt != NULL)
    {
        if(!strstr(pDirEnt->d_name, ".")) // if there is no dot in the name
        {
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , next_hue()
                        , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                        , 0
                        , font_size_factor
                        , "%s                  "
                        , pDirEnt->d_name
                      );
            y_off += y_line;
            found_one = true;
        }
        pDirEnt = readdir(pDIR);
    }
    closedir(pDIR);
    if(!found_one)
    {
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "none found        "
                  );
        y_off += y_line;
    }
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::apply_wave_offsets(string file_name)
{
    int                  i                           ,
                         span                        ,
                         last                        ,
                         sample_count       = 0      ,
                         offset_in_max               ,
                         offset_in_min               ,
                         offset_applied_max          ,
                         offset_applied_min          ,
                        *offset_applied     =  NULL  ;
    bool                 inverted           =  false ;
    ifstream             wave_ifstream               ;
    ofstream             wave_ofstream               ;
    ez_wave_header       wave_in_header              ,
                         wave_out_header             ;                         
    ez_wave_sample       wave_sample(6)              ;
    ez_wave_sample_pointer *sample_window            ;
    //------------------------------------------------------------------------
    process_exception = false;
    //------------------------------------------------------------------------
    remove((EZ_ILDA_WAVE_SHARE + "__TEMP__.wav").c_str());
    //------------------------------------------------------------------------
    if(    file_name.size() < 4
        || file_name.substr(file_name.size() - 4) != ".wav"
      )
        file_name += ".wav";
    //------------------------------------------------------------------------
    if(rename(   (EZ_ILDA_WAVE_SHARE + file_name).c_str()
               , (EZ_ILDA_WAVE_SHARE + "__TEMP__.wav").c_str()
             )
      )
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , BMP.white
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , (EZ_ILDA_WAVE_SHARE + file_name + " failed to remane      ").c_str()
                  );
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    wave_ifstream.open((EZ_ILDA_WAVE_SHARE + "__TEMP__.wav").c_str(), ios::in | ios::binary);
    if(!wave_ifstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , BMP.white
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "[any key] %s failed to open          "
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    if(wave_in_header.from_ifstream(wave_ifstream))
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , BMP.white
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "[any key] error reading %s wave header"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    if(wave_in_header.num_channels < 6)
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , BMP.white
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "[any key] %s is not a 6 channel wave"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    wave_ofstream.open((EZ_ILDA_WAVE_SHARE + file_name).c_str(), ios::out | ios::binary);
    if(!wave_ofstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , BMP.white
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "[any key] %s"
                    , (EZ_ILDA_WAVE_SHARE + file_name + " failed to open for output").c_str()
                  );
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    wave_out_header = wave_in_header;
    wave_out_header.ez_wave_mode = wave_out_header.ez_wave_mode | EZ_WAVE_OFFSETS;
    offset_applied  = new int[6];
    //------------------------------------------------------------------------
    for(i = 0; i < 6; i++)
        offset_applied[i] = wave_in_header.offset[i] - wave_offsets[i];
    //------------------------------------------------------------------------
    offset_in_min      = wave_offsets[0];
    offset_in_max      = wave_offsets[0];
    offset_applied_min = offset_applied[0];
    offset_applied_max = offset_applied[0];
    //------------------------------------------------------------------------
    for(i = 1; i < 6; i++)
    {
        if(wave_offsets[i] > offset_in_max)
            offset_in_max = wave_offsets[i];
        if(wave_offsets[i] < offset_in_min)
            offset_in_min = wave_offsets[i];

        if(offset_applied[i] > offset_applied_max)
            offset_applied_max = offset_applied[i];
        if(offset_applied[i] < offset_applied_min)
            offset_applied_min = offset_applied[i];
    }    
    //------------------------------------------------------------------------
    span = (offset_applied_max - offset_applied_min) + 1;
    last = span - 1;
    //------------------------------------------------------------------------
    if(last == 0) // no shift
    {
        wave_out_header.to_ofstream(wave_ofstream);        
        while(true)
        {
            if(wave_sample.from_ifstream(wave_ifstream)) // read another sample from file until EOF
                break;            
            wave_sample.to_ofstream(wave_ofstream);
            sample_count++;
        }
    }
    else
    {
        //--------------------------------------------------------------------
        for(i = 0; i < 6; i++)
        {
            wave_out_header.offset[i] = -(offset_in_max - wave_offsets[i]);
            offset_applied[i] = offset_applied_max - offset_applied[i];
        }
        //--------------------------------------------------------------------
        sample_window = new ez_wave_sample_pointer[span];
        for(i = 0; i < span; i++)
            sample_window[i] = new ez_wave_sample(6); // span element array of (6 channel ez_wave_samples)
        //--------------------------------------------------------------------
        wave_out_header.num_samples  = wave_in_header.num_samples + 1;
        wave_out_header.to_ofstream(wave_ofstream);
        //--------------------------------------------------------------------
        while(true)
        {
            for(i = 0; i < last; i++)
                sample_window[i] = sample_window[i + 1];            
            for(i = 0; i < 6; i++) // each channel
                wave_sample.channel[i] = sample_window[offset_applied[i]]->channel[i];
            wave_sample.to_ofstream(wave_ofstream); // write it to disk
            if(sample_window[last]->from_ifstream(wave_ifstream)) // read another sample from file until EOF
                break;            
            sample_count++;
        }
//        for(i = 0; i < span; i++)
//            delete sample_window[i];
        delete sample_window;
    }
    wave_ifstream.close();
    wave_ofstream.close();
    remove((EZ_ILDA_WAVE_SHARE + "__TEMP__.wav").c_str());
    delete offset_applied;
    return;    
}

//############################################################################
void ez_ilda::invert_wave(string file_name)
{
    int                  i                ,
                         sample_count  = 0;
    ifstream             wave_ifstream    ;
    ofstream             wave_ofstream    ;
    ez_wave_header       wave_in_header   ,
                         wave_out_header  ;
    ez_wave_sample       wave_sample      ;
    //------------------------------------------------------------------------
    process_exception = false;
    //------------------------------------------------------------------------    
    remove((EZ_ILDA_WAVE_SHARE + "__TEMP__.wav").c_str());
    //------------------------------------------------------------------------
    if(    file_name.size() < 4
        || file_name.substr(file_name.size() - 4) != ".wav"
      )
        file_name += ".wav";
    //------------------------------------------------------------------------
    if(rename(   (EZ_ILDA_WAVE_SHARE + file_name).c_str()
               , (EZ_ILDA_WAVE_SHARE + "__TEMP__.wav").c_str()
             )
      )
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , BMP.white
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , (EZ_ILDA_WAVE_SHARE + file_name + " failed to remane      ").c_str()
                  );
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    wave_ifstream.open((EZ_ILDA_WAVE_SHARE + "__TEMP__.wav").c_str(), ios::in | ios::binary);
    if(!wave_ifstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , BMP.white
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "[any key] %s failed to open          "
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    if(wave_in_header.from_ifstream(wave_ifstream))
    {
        wave_ifstream.close();                                                   
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , BMP.white
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "[any key] error reading %s wave header"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    wave_ofstream.open((EZ_ILDA_WAVE_SHARE + file_name).c_str(), ios::out | ios::binary);
    if(!wave_ofstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * font_size_factor
                    , BMP.white
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "[any key] %s"
                    , (EZ_ILDA_WAVE_SHARE + file_name + " failed to open for output").c_str()
                  );
        bmp_to_SDL_Surface(&bmp, screen);
        process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    wave_out_header = wave_in_header;
    if(wave_in_header.ez_wave_mode & EZ_WAVE_POSITIVE)
        wave_out_header.ez_wave_mode = wave_out_header.ez_wave_mode & ~EZ_WAVE_POSITIVE;
    else
        wave_out_header.ez_wave_mode = wave_out_header.ez_wave_mode | EZ_WAVE_POSITIVE;
    wave_out_header.to_ofstream(wave_ofstream);
    while(true)
    {
        if(wave_sample.from_ifstream_inverted(wave_ifstream)) // EOF
            break;
        wave_sample.to_ofstream(wave_ofstream);
    }    
    wave_ifstream.close();
    wave_ofstream.close();
    remove((EZ_ILDA_WAVE_SHARE + "__TEMP__.wav").c_str());    
    return;
}

//############################################################################
void ez_ilda::load_background_bitmap(struct ezfb_bitmap* bmp)
{
    bmp_flip(bmp);
    bmp_into_bmp(&BACKGROUND, bmp);
    return;
}

//############################################################################
void ez_ilda::load_background_bitmap(string file)
{
    char file_name[81];
    struct ezfb_bitmap bmp;
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".bmp"
      )
        file += ".bmp";
    strcpy(file_name, (EZ_ILDA_BMP_SHARE + file).c_str());
    if(bmp_from_file(&bmp, file_name))
    {
        load_background_bitmap(&bmp);
        bmp_free(&bmp);
        bg_file_name = file;
    }
    return;
}

//############################################################################
void ez_ilda::clear_background_bitmap()
{
    bmp_uint(&BACKGROUND, bg_color.as_uint());
    bg_file_name = ".";
    return;
}

//############################################################################
void ez_ilda::recolor_background()
{
    bmp_uint(&BACKGROUND, bg_color.as_uint());
    if(bg_file_name != ".")
        load_background_bitmap(bg_file_name);
    return;
}

//############################################################################
void ez_ilda::display_frame_effects()
{
    int i,
        y_line = 8 * font_size_factor,
        x_off  = BMP.yres + EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor;
    
    bmp_printf(   &BMP
                , x_off
                , 0
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "frame effects"
              );
    if(frame_effect_names.size() > 0)
    {
        for(i = 0; i < (int)frame_effect_names.size(); i++)
        {
            y_off += y_line;
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , next_hue()
                        , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                        , 0
                        , font_size_factor
                        , "%s          "
                        , frame_effect_names[i].c_str()
                      );
        }
    }
    else
    {
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "no frame effects loaded"
                  );
    }        
    bmp_to_SDL_Surface(&BMP, screen);
    return;
}

//############################################################################
void ez_ilda::display_frame_set_effects()
{
    int i,
        y_line = 8 * font_size_factor,
        x_off  = BMP.yres + EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor;
    
    bmp_printf(   &BMP
                , x_off
                , 0
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                , 0
                , font_size_factor
                , "frame set effects"
              );
    if(frame_set_effect_names.size() > 0)
    {
        for(i = 0; i < (int)frame_set_effect_names.size(); i++)
        {
            y_off += y_line;
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , next_hue()
                        , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                        , 0
                        , font_size_factor
                        , "%s          "
                        , frame_set_effect_names[i].c_str()
                      );
        }
    }
    else
    {
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.make_rgb(bg_color.r, bg_color.g, bg_color.b)
                    , 0
                    , font_size_factor
                    , "no frame set effects loaded"
                  );
    }        
    bmp_to_SDL_Surface(&BMP, screen);
    return;
}

//############################################################################
void ez_ilda::call_frame_effect(string name)
{
    int i;
    for(i = 0; i < (int)frame_effect_names.size(); i++)
        if(frame_effect_names[i] == name)
            break;
    if(i < (int)frame_effect_names.size())
        frame_effects[i](&(current_frame()));
    return;
}

//############################################################################
void ez_ilda::call_frame_set_effect(string name)
{
    int i;
    for(i = 0; i < (int)frame_set_effect_names.size(); i++)
        if(frame_set_effect_names[i] == name)
            break; 
    if(i < (int)frame_set_effect_names.size())
        frame_set_effects[i](this);
    return;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
